#version 330 core
uniform float size;

layout (points) in;
layout (triangle_strip, max_vertices = 4) out;
    
uniform mat4 projection;
uniform mat4 view;
uniform mat4 model;

in VertexData {
    vec4 ex_texCoord;
    float ex_color;
} VertexIn[1];


out VertexData {
    vec4 ex_texCoord;
} VertexOut;

void main()                                                   
{                                                             
	// aging                                                  
	//float phase = VertexIn[0].ex_texCoord.w;    
	//float phase = 0.5; 
	float phase = VertexIn[0].ex_color;
	float radius = size;

	// eye space
	vec3 pos = (view * model * gl_in[0].gl_Position).xyz;
	vec3 pos2 = VertexIn[0].ex_texCoord.xyz;

	vec3 motion = pos - pos2;
	vec3 dir = normalize(motion);
	float len = length(motion);

	vec3 x = dir * radius;
	vec3 view = normalize(-pos);
	vec3 y = normalize(cross(dir, view)) * radius;
	float facing = dot(view, dir);

	// check for very small motion to avoid jitter
	float threshold = 0.01;
	if ((len < threshold) || (facing > 0.95) || (facing < -0.95)) 
	{
        	pos2 = pos;
	        x = vec3(radius, 0.0, 0.0);
        	y = vec3(0.0, -radius, 0.0);
	}

	// output quad
	mat4 transformation = projection;
	
	VertexOut.ex_texCoord = vec4(0.0, 0.0, 0.0, phase);
	gl_Position = transformation * vec4(pos - x - y, 1.0);
	EmitVertex();

	VertexOut.ex_texCoord = vec4(0.0, 1.0, 0.0, phase);
	gl_Position = transformation * vec4(pos - x + y, 1.0);
	EmitVertex();

	VertexOut.ex_texCoord = vec4(1.0, 0.0, 0.0, phase);                  
	gl_Position = transformation * vec4(pos + x - y, 1.0);
    	EmitVertex();

	VertexOut.ex_texCoord = vec4(1.0, 1.0, 0.0, phase);
	gl_Position =  transformation * vec4(pos + x + y, 1.0);
	EmitVertex();
}  