#version 330 core

// uniform shader-parameters
uniform sampler2D img;
uniform float exposure; // 
uniform float brightMax;
uniform sampler1D average;
uniform int mode;
uniform int enable;

// fragment-shader input variables
in vec2 ex_texcoord;

// fragment-shader output variable (-> stored in the frame-buffer, i.e. "the pixel you see")
out vec4 out_color;


void main()
{    
	vec4 color = texture(img, ex_texcoord);

	if(enable > 0)
	{
	//textureLod(img, vec2(0.5,0.5), 9);
	//float averageY = (texture(average, (0.5,0.5))).x;
	vec4 averageYvec = clamp((textureLod(img, vec2(0.5,0.5), 9)), vec4(0.2) , vec4(0.9));
	float averageY = dot(vec4(0.30, 0.59, 0.11, 0.0), averageYvec);
	float a = max(0, (1.5 - (1.5 / (averageY*0.1 + 1.0)))) + 0.1; // 
	float maxY = 2.5;
	
	// Perform tone-mapping
	float Y = dot(vec4(0.30, 0.59, 0.11, 0.0), color); // luminance of a pixel
	float scaleY = a * (Y / averageY); // linear mapping
	float reinhard = scaleY / (1.0 + scaleY); // Reinhard's operator
	float YD = scaleY * (1.0 + (scaleY / pow(maxY,2))) / (scaleY + 1.0);// modified Reinhards operator
	//float YD = exposure * (exposure/brightMax + 1.0) / (exposure + 1.0); 
	
	float adaptivLog = (a / (log(averageY + 1.0)/ log(10.0))) * (log(scaleY + 1.0) / log( 2.0 + ( pow((scaleY/averageY),(log(0.7)/log(0.5))) * 8)));
	
	if(mode == 0) color *= scaleY;//linear
	if(mode == 1) color *= reinhard;//reinhard
	if(mode == 2) color *= YD;//mod reinhard
	if(mode == 3) color *= adaptivLog;//log
	
	//out_color = vec4(averageY,averageY,averageY,1.0);
	}
	
	
	out_color = color;
}

