#version 330 core  
precision mediump float;

// sampler holding texture data
uniform int transfer;
uniform int mip;
uniform int average;
uniform int light;
uniform int phong;
uniform int adaptive;
uniform float stepSize;
uniform sampler2D entryTexture;
uniform sampler2D exitTexture;
uniform sampler3D colorTexture;
uniform sampler3D pcdTexture;
uniform sampler1D transferTexture;
uniform sampler1D tex_hg;
uniform sampler1D tex_hg_dev;

uniform mat4 perspective;
uniform mat4 view;
uniform mat4 model;

// fragment-shader input variables
in vec2 fragTexCoordVolume;

// fragment-shader output variable (-> stored in the frame-buffer, i.e. "the pixel you see")
out vec4 out_color;


//phong lightening
vec3 lightening(vec3 n, vec3 viewDir, vec3 color)
{
	vec3 normal = normalize(n);
	vec3 lightColor = vec3(1.0, 1.0, 1.0);
	vec3 ambientColor = vec3(0.01, 0.1, 0.1);
	vec3 diffuseColor = vec3(0.8, 0.8, 0.8);
	
	vec3 lightDir = viewDir;
	vec3 highlightDir = normalize(lightDir + viewDir);
	float specExponent = 50; //je groesser der exponent desto kleiner der glanzpunkt desto glatter ist die oberflche
	
	vec3 result;
	vec3 ambient = ambientColor * color;
	vec3 diffuse = diffuseColor * min(1.0, max(0.0, dot(normal, lightDir))) * color;
	
	vec3 specular = 0.5 * lightColor * pow(max(0.0, dot(normal, highlightDir)), specExponent) * color;
	//vec3 specular = color;
	result = min(vec3(1.0), max(vec3(0.0), ambient + diffuse + specular));
	return result;
/*/
	return abs(normal);
//*/
}
vec3 calculateGradients(vec3 point, float dz)
{
	return normalize(texture(colorTexture, point.xyz).xyz);	
}

void maximumIntensity(in float dz, in vec4 entry_point, in vec4 exit_point, in int maxiter, in vec3 diff, out vec3 pos, out vec4 outcolor)
{
	vec3 point = entry_point.xyz;
	vec3 N;
	vec3 V;
	
	int dist = maxiter+1;
	
	if (entry_point.w == 0.0) {
		discard; //alpha = 0 man sieht nichts!
	} else {
		vec4 Result = vec4(0, 0, 0, 0);
		for (int i = 0; i < maxiter; i++) {

		vec4 intensity = vec4(texture(colorTexture, point).w);
			
			if (intensity.w > Result.w) {
				Result = vec4(intensity.w);
				
				if (transfer == 1) {
					vec4 transferred = texture(transferTexture, (Result.w));
					Result.xyz = transferred.xyz;
				}
			
				if (light == 1) {
					N = calculateGradients(point, dz);
					V = vec3(exit_point - entry_point);
					Result.xyz = lightening(normalize(N), V, Result.xyz);
				}
			}
			if(Result.w > 0.2 && i < dist) 
			{
				pos = point;
				dist = i;
			}
			point += diff;
	
			if (Result.w >= 1.0) {
				break;
			}
				
		}

		outcolor = Result * Result.w;
		return;
	}
}

void main()
{
	float dz = stepSize; //Schrittgre = 1/200
	vec4 entry_point = texture(entryTexture, fragTexCoordVolume);
	vec4 exit_point = texture(exitTexture, fragTexCoordVolume);
	float iterations = distance(entry_point, exit_point)/dz;
	vec3 diff = (exit_point.xyz - entry_point.xyz)/iterations;
	int maxiter = int(floor(iterations));

	float p = 1.0;
	vec3 pos;
	
	maximumIntensity(dz, entry_point, exit_point, maxiter, diff, pos, out_color);
	
	if( length(exit_point.xyz-entry_point.xyz) > 0.01)
	{
		pos -= 0.5;
		pos *= 2.0;
		pos -= vec3(0.0, 0.0, 2.0);
		vec4 p2 = perspective * view * vec4(pos.x, pos.y, pos.z, 1.0);
		p = p2.z/p2.w;
		p = (p*0.5)+0.5;
	}
	gl_FragDepth = p;
}