#version 330 core
uniform float size;

layout (points) in;
layout (triangle_strip, max_vertices = 4) out;
    
uniform mat4 projection;
uniform mat4 view;
uniform mat4 model;

uniform sampler3D vol;

in VertexData {
    vec4 ex_texCoord;
    float ex_color;
} VertexIn[1];


out VertexData {
    vec4 ex_texCoord;
	flat int frontface;
} VertexOut;

mat4 PopulateMatrix(vec3 xCol, vec3 yCol, vec3 zCol) 
{
	mat4 res = mat4(1.0);

	res[0][0] = xCol.x;
    res[1][0] = xCol.y;
    res[2][0] = xCol.z;

    res[0][1] = yCol.x;
    res[1][1] = yCol.y;
    res[2][1] = yCol.z;

    res[0][2] = zCol.x;
    res[1][2] = zCol.y;
	res[2][2] = zCol.z;

	return transpose(res);
}




// Convert from Euler Angles
vec4 FromEuler(float pitch, float yaw, float roll)
{
	// Basically we create 3 Quaternions, one for pitch, one for yaw, one for roll
	// and multiply those together.
	// the calculation below does the same, just shorter
 
	float PIOVER180 = 3.14159265 / 180.0;

	/*
	float p = pitch * PIOVER180 / 2.0;
	float y = yaw * PIOVER180 / 2.0;
	float r = roll * PIOVER180 / 2.0;
	/*/
	float p = pitch;
	float y = yaw;
	float r = roll;
 
	float sinp = sin(p);
	float siny = sin(y);
	float sinr = sin(r);
	float cosp = cos(p);
	float cosy = cos(y);
	float cosr = cos(r);
 
	vec4 quat;
	quat.x = sinr * cosp * cosy - cosr * sinp * siny;
	quat.y = cosr * sinp * cosy + sinr * cosp * siny;
	quat.z = cosr * cosp * siny - sinr * sinp * cosy;
	quat.w = cosr * cosp * cosy + sinr * sinp * siny;
 
	quat = normalize(quat);

	return quat;
}

// Convert to Matrix
mat4 getMatrix(vec4 quat)
{
	float x = quat.x;
	float y = quat.y;
	float z = quat.z;
	float w = quat.w;

	float x2 = x * x;
	float y2 = y * y;
	float z2 = z * z;
	float xy = x * y;
	float xz = x * z;
	float yz = y * z;
	float wx = w * x;
	float wy = w * y;
	float wz = w * z;
 
	// This calculation would be a lot more complicated for non-unit length quaternions
	// Note: The constructor of Matrix4 expects the Matrix in column-major format like expected by
	//   OpenGL
	return mat4( 1.0f - 2.0f * (y2 + z2), 2.0f * (xy - wz), 2.0f * (xz + wy), 0.0f,
				 2.0f * (xy + wz), 1.0f - 2.0f * (x2 + z2), 2.0f * (yz - wx), 0.0f,
				 2.0f * (xz - wy), 2.0f * (yz + wx), 1.0f - 2.0f * (x2 + y2), 0.0f,
				 0.0f, 0.0f, 0.0f, 1.0f);
}

vec4 fromAngleAxis(float angle, vec3 axis)
{
	float fHalfAngle = 0.5f*angle;
	float fSin = sin(fHalfAngle);

	vec4 q;
	q.w = cos(fHalfAngle);
	q.x = fSin*axis.x;
	q.y = fSin*axis.y;
	q.z = fSin*axis.z;
	return q;
}

//http://books.google.at/books?id=hiBFUv_FT0wC&pg=PA214&lpg=PA214&dq=Stan+Melax%27s+article+in+Game+Programming+Gems&source=bl&ots=OCeLOwDdYl&sig=1dSf_TzbqsBBXQm3I43X1A_kP6Y&hl=en&sa=X&ei=-p_cT5rZNseQswaov-HxDQ&redir_esc=y#v=onepage&q&f=false 
vec4 rotateTo(vec3 from, vec3 to)
{
	float x = 0.0;
	float y = 0.0;
	float z = 0.0; 
	vec4 q;

	// Based on Stan Melax's article in Game Programming Gems
	vec3 v0 = from;
	vec3 v1 = to;
	v0 = normalize(v0);
	v1 = normalize(v1);

	float d = dot(v0,v1);
	if (d >= 1.0f)	return vec4(0.0,0.0,0.0,1.0);
	float s = sqrt( (1.0+d)*2.0 );
	vec3 c = cross(v0, v1);

	q.x = c.x / s;
	q.y = c.y / s;
	q.z = c.z / s;
	q.w = s * 0.5f;
 
	return q;
}

void main()                                                   
{                 	                                 
	// aging                                                  
	float phase = VertexIn[0].ex_color;
	float radius = size*5.0;

	/*
	// world space
	vec3 pos = (gl_in[0].gl_Position).xyz;
	vec3 velocity = normalize(VertexIn[0].ex_texCoord.xyz);

	//Quaternion q;
	//vector a = crossproduct(v1, v2)
	//q.xyz = a;
	//q.w = sqrt((v1.Length ^ 2) * (v2.Length ^ 2)) + dotproduct(v1, v2)

	//Don't forget to normalize q
	//Richard is right about there not being a unique rotation, but the above should give the "shortest arc". Which is probably what you need.

	vec3 v1 = vec3(0.0, 1.0, 0.0);
	
	vec4 q;
	vec3 a = cross(v1, velocity);
	q.xyz = a;
	q.w = sqrt(pow(size*2,2.0)+pow(size*2,2.0)) + dot(v1, velocity);

	q = normalize(q);
	mat4 rot = transpose(getMatrix(q));

	// output quad
	mat4 transformation = projection*view*model;
	
	VertexOut.ex_texCoord = vec4(0.0, 0.0, 0.0, phase);
	gl_Position = transformation * vec4(pos + (rot*vec4(0,-radius, 0.0, 1.0)).xyz, 1.0);
	EmitVertex();

	VertexOut.ex_texCoord = vec4(0.0, 1.0, 0.0, phase);
	gl_Position = transformation * vec4(pos + (rot*vec4(-radius, 0.0, 0.0, 1.0)).xyz, 1.0);
	EmitVertex();

	VertexOut.ex_texCoord = vec4(1.0, 1.0, 0.0, phase);
	gl_Position = transformation * vec4(pos + (rot*vec4( radius, 0.0, 0.0, 1.0)).xyz, 1.0);
	EmitVertex();

	VertexOut.ex_texCoord = vec4(1.0, 0.0, 0.0, phase);                  
	gl_Position = transformation * vec4(pos + (rot*vec4( 0.0, radius, 0.0, 1.0)).xyz, 1.0);
	EmitVertex();
	*/

	
	vec3 norm = normalize(texture(vol, gl_in[0].gl_Position.xyz).xyz);
	vec3 tan = normalize(VertexIn[0].ex_texCoord.xyz);
	vec3 binorm = normalize(cross(norm,tan));
	norm *= radius;
	tan *= radius;
	binorm *= radius;

	vec4 pos = vec4(gl_in[0].gl_Position.xyz,1.0);

	// output quad
	mat4 transformation = projection*view;

	vec4 a = transformation * vec4((model*pos).xyz + tan, 1.0);
	vec4 b = transformation * vec4((model*pos).xyz + binorm,1.0);
	vec4 c = transformation * vec4((model*pos).xyz - binorm,1.0);
	vec4 d = transformation * vec4((model*pos).xyz - tan, 1.0);
	
	int fface = 1;
	float ftmp = dot(b-a, c-a);
	if(ftmp < 0.0) fface = 0;

	VertexOut.ex_texCoord = vec4(1.0, 1.0, 0.0, phase);
	VertexOut.frontface = fface;
	gl_Position = a;
	EmitVertex();

	VertexOut.ex_texCoord = vec4(1.0, 0.0, 0.0, phase);
	VertexOut.frontface = fface;
	gl_Position = b;
	EmitVertex();

	VertexOut.ex_texCoord = vec4(0.0, 1.0, 0.0, phase);
	VertexOut.frontface = fface;
	gl_Position = c;
	EmitVertex();

	VertexOut.ex_texCoord = vec4(0.0,0.0, 0.0, phase);                  
	VertexOut.frontface = fface;
	gl_Position = d;
	EmitVertex();
}  