#version 330 core  
precision mediump float;

// sampler holding texture data
uniform int transfer;
uniform int mip;
uniform int average;
uniform int light;
uniform int phong;
uniform int adaptive;
uniform float stepSize;
uniform sampler2D entryTexture;
uniform sampler2D exitTexture;
uniform sampler3D colorTexture;
uniform sampler1D transferTexture;
uniform sampler1D tex_hg;
uniform sampler1D tex_hg_dev;

uniform mat4 view;

// fragment-shader input variables
in vec2 fragTexCoordVolume;

// fragment-shader output variable (-> stored in the frame-buffer, i.e. "the pixel you see")
out vec4 out_color;


vec4 lerp(vec4 a, vec4 b, float w)
{
  return a + w*(b-a);	
//	return (a*(1.0-w)) + (w*b);
}
vec3 lerp(vec3 a, vec3 b, float w)
{
  return a + w*(b-a);
//	return (a*(1.0-w)) + (w*b);
}
vec2 lerp(vec2 a, vec2 b, float w)
{
  return a + w*(b-a);
//	return (a*(1.0-w)) + (w*b);
}
float lerp(float a, float b, float w)
{
  return a + w*(b-a);
//return (a*(1.0-w)) + (w*b);
}

//coord = texture coordinate
//size_source = 3d Texture Gre
float calculateVoxel(vec3 coord, vec3 size_source )
{
	//kann dann 1zu1 fr cuda genommen werden (beides bentzt normalisierte Texturekoordinaten)
	// calc filter texture coordinates where [0,1] is a single texel
	// (can be done in vertex program instead)
	//Es ist ja nur der Nachkommabereich interessant von den Texturkoordinaten [0,128]
	//Nachdem ich GL_REPEAT bei der Textur habe, sollte es passen.

vec3 coord_hg = (coord*size_source) - vec3(0.5f, 0.5f, 0.5f);
//vec3 coord_hg = (coord) - vec3(0.5f, 0.5f, 0.5f)/size_source;

	// fetch offsets and weights from filter texture
	vec4 hg_x = texture( tex_hg, coord_hg.x );
	vec4 hg_y = texture( tex_hg, coord_hg.y );
	vec4 hg_z = texture( tex_hg, coord_hg.z );

	// determine linear sampling coordinates
	//Scheint zu stimmen: 
	//Wie funktioniert das:
	//coord_sourceXYZ, wobei X,Y,Z ist {0,1}
	//bei A = 0.an Stelle Z.. gehrt: coord_sourceXY0 = coord_sourceXY0 - hg_Z.y / size_source.Z,
	//bei A = 1 an Stelle Z... gehrt: coord_sourceXY1 = coord_sourceXY0 + hg_Z.x / size_source.Z,
	//, wobei A ist X, Y oder Z von coord_sourceXYZ.

	vec3 coord_source100 = coord - (hg_x.x/size_source.x);
	vec3 coord_source000 = coord + (hg_x.y/size_source.x);

	vec3 coord_source110 = coord_source100 - (hg_y.x/size_source.y);
	vec3 coord_source010 = coord_source000 - (hg_y.x/size_source.y);
	     coord_source100 = coord_source100 + (hg_y.y/size_source.y);
	     coord_source000 = coord_source000 + (hg_y.y/size_source.y);

	vec3 coord_source111 = coord_source110 - (hg_z.x/size_source.z);
	vec3 coord_source011 = coord_source010 - (hg_z.x/size_source.z);
	vec3 coord_source101 = coord_source100 - (hg_z.x/size_source.z);
	vec3 coord_source001 = coord_source000 - (hg_z.x/size_source.z);
	     coord_source110 = coord_source110 + (hg_z.y/size_source.z);
	     coord_source010 = coord_source010 + (hg_z.y/size_source.z);
	     coord_source100 = coord_source100 + (hg_z.y/size_source.z);
	     coord_source000 = coord_source000 + (hg_z.y/size_source.z);


	//ab da sollte es wieder stimmen
	// fetch four linearly interpolated inputs
	vec4 n_val;
	vec4 f_val;
	f_val.w = texture( colorTexture, coord_source000).w;
	f_val.z = texture( colorTexture, coord_source100).w;
	f_val.y = texture( colorTexture, coord_source010).w;
	f_val.x = texture( colorTexture, coord_source110).w;
	n_val.w = texture( colorTexture, coord_source001).w;
	n_val.z = texture( colorTexture, coord_source101).w;
	n_val.y = texture( colorTexture, coord_source011).w;
	n_val.x = texture( colorTexture, coord_source111).w;


	// weight along z direction
	//Das stimmt einmal, da g0 in z gespeichert ist
	//Daher (tex_source000*(1.0-hg_z.z)) + (tex_source001*hg_z.z)

	f_val = lerp(f_val, n_val, hg_z.z );
	f_val.xy = lerp( f_val.xy, f_val.zw, hg_y.z );
	return lerp( f_val.w, f_val.z, hg_x.z );
}

float calcGrad(vec4 hg_x, vec4 hg_y, vec4 hg_z, int type, vec3 coord, vec3 size_source)
{
	float h = (hg_x.x/size_source.x);
	vec3 coord_source100 = coord - h;
	vec3 coord_source000 = coord + h;

	h = (hg_y.x/size_source.y);
	vec3 coord_source110 = coord_source100 - h;
	vec3 coord_source010 = coord_source000 - h;
	     coord_source100 = coord_source100 + h;
	     coord_source000 = coord_source000 + h;

	h = (hg_z.x/size_source.z);
	vec3 coord_source111 = coord_source110 - h;
	vec3 coord_source011 = coord_source010 - h;
	vec3 coord_source101 = coord_source100 - h;
	vec3 coord_source001 = coord_source000 - h;
	     coord_source110 = coord_source110 + h;
	     coord_source010 = coord_source010 + h;
	     coord_source100 = coord_source100 + h;
	     coord_source000 = coord_source000 + h;


	//ab da sollte es wieder stimmen
	// fetch four linearly interpolated inputs
	vec4 n_val;
	vec4 f_val;
	f_val.w = texture( colorTexture, coord_source000).w;
	f_val.z = texture( colorTexture, coord_source100).w;
	f_val.y = texture( colorTexture, coord_source010).w;
	f_val.x = texture( colorTexture, coord_source110).w;
	n_val.w = texture( colorTexture, coord_source001).w;
	n_val.z = texture( colorTexture, coord_source101).w;
	n_val.y = texture( colorTexture, coord_source011).w;
	n_val.x = texture( colorTexture, coord_source111).w;


	float gradient = 0.0;

	// weight along z direction	
	switch(type)
	{
	case 0: f_val = lerp(n_val, f_val, hg_z.z);
		f_val.zw = lerp(f_val.xy, f_val.zw, hg_y.z);
		gradient = (f_val.w - f_val.z)*hg_x.z;
	break;
	case 1:	f_val = lerp(n_val, f_val, hg_z.z);	
		f_val.zw = (f_val.zw-f_val.xy)*hg_y.z;		
		gradient = lerp(f_val.z, f_val.w, hg_x.z);
	break;
	case 2:	f_val = (f_val-n_val)*hg_z.z;
		f_val.zw = lerp(f_val.xy, f_val.zw, hg_y.z);
		gradient = lerp(f_val.z, f_val.w, hg_x.z);	
	break;
	}

	return abs(gradient);
}

//VOm Assembercode...
vec3 calculateGradient(vec3 coord, vec3 size_source )
{
	size_source  =  vec3(128.0);

	//kann dann 1zu1 fr cuda genommen werden (beides bentzt normalisierte Texturekoordinaten)
	// calc filter texture coordinates where [0,1] is a single texel
	// (can be done in vertex program instead)
	//Es ist ja nur der Nachkommabereich interessant von den Texturkoordinaten [0,128]
	//Nachdem ich GL_REPEAT bei der Textur habe, sollte es passen.

vec3 coord_hg = (coord*size_source) - vec3(0.5f, 0.5f, 0.5f);
//vec3 coord_hg = (coord) - vec3(0.5f, 0.5f, 0.5f)/size_source;

	// fetch offsets and weights from filter texture
	vec4 hg_x = texture( tex_hg, coord_hg.x);
	vec4 hg_y = texture( tex_hg, coord_hg.y);
	vec4 hg_z = texture( tex_hg, coord_hg.z);

	vec4 hg_x_dev = texture( tex_hg_dev, coord_hg.x);
	vec4 hg_y_dev = texture( tex_hg_dev, coord_hg.y);
	vec4 hg_z_dev = texture( tex_hg_dev, coord_hg.z);

	vec3 gradient;
	gradient.x = calcGrad(hg_x_dev, hg_y, hg_z, 0, coord, size_source);
	gradient.y = calcGrad(hg_x, hg_y_dev, hg_z, 1, coord, size_source);
	gradient.z = calcGrad(hg_x, hg_y, hg_z_dev, 2, coord, size_source);

//gradient = (view * vec4(gradient, 0.0)).xyz;

	return (gradient);
}



//phong lightening
vec3 lightening(vec3 normal, vec3 viewDir, vec3 color)
{
	vec3 lightColor = vec3(1.0, 1.0, 1.0);
	vec3 ambientColor = vec3(0.01, 0.1, 0.1);
	vec3 diffuseColor = vec3(0.8, 0.8, 0.8);
	
	vec3 lightDir = viewDir;
	vec3 highlightDir = normalize(lightDir + viewDir);
	float specExponent = 50; //je groesser der exponent desto kleiner der glanzpunkt desto glatter ist die oberflche
	
	vec3 ambient = ambientColor * color;
	vec3 diffuse = diffuseColor * max(0.0, dot(normal, lightDir)) * color;
	
	vec3 result;
	if (phong == 1) {
		vec3 specular = 0.5 * lightColor * pow(max(0.0, dot(normal, highlightDir)), specExponent) * color;
		result = ambient + diffuse + specular;
	} else {
		result = ambient + diffuse;
	}
	return result;
/*/
	return abs(normal);
//*/
}

vec3 calculateGradients(vec3 point, float dz)
{
/*
	dz = 1.0/128.0;

	vec3 sample1 = vec3(0, 0, 0);
	vec3 sample2 = vec3(0, 0, 0);
	
	sample1.x = texture(colorTexture, point.xyz - vec3(dz, 0, 0)).w;
	sample1.y = texture(colorTexture, point.xyz - vec3(0, dz, 0)).w;
	sample1.z = texture(colorTexture, point.xyz - vec3(0, 0, dz)).w;
	
	sample2.x = texture(colorTexture, point.xyz + vec3(dz, 0, 0)).w;
	sample2.y = texture(colorTexture, point.xyz + vec3(0, dz, 0)).w;
	sample2.z = texture(colorTexture, point.xyz + vec3(0, 0, dz)).w;
	
	vec3 N = sample2 - sample1;
	
	return N;
/*/

	/*
	//Gradienten durch Aufruf von calculateVoxel
	dz = 1.0/128.0;

	vec3 sample1 = vec3(0, 0, 0);
	vec3 sample2 = vec3(0, 0, 0);
	
	sample1.x = calculateVoxel(point.xyz - vec3(dz, 0, 0), vec3(128.0));
	sample1.y = calculateVoxel(point.xyz - vec3(0, dz, 0), vec3(128.0));
	sample1.z = calculateVoxel(point.xyz - vec3(0, 0, dz), vec3(128.0));
	
	sample2.x = calculateVoxel(point.xyz + vec3(dz, 0, 0), vec3(128.0));
	sample2.y = calculateVoxel(point.xyz + vec3(0, dz, 0), vec3(128.0));
	sample2.z = calculateVoxel(point.xyz + vec3(0, 0, dz), vec3(128.0));
	
	vec3 N = sample2 - sample1;
	
	return N;
	/*/
	//Gradienten durch Normalverteilung

		
		return normalize(calculateGradient(point.xyz, vec3(128.0)));	
		/*/
		return normalize(texture(colorTexture, point.xyz).xyz);
		//*/

	//*/
//*/
}

vec4 maximumIntensity(float dz, vec4 entry_point, vec4 exit_point, int maxiter, vec3 diff)
{
	vec3 point = entry_point.xyz;
	vec3 N;
	vec3 V;
	
	if (entry_point.w == 0.0) {
		discard; //alpha = 0 man sieht nichts!
	} else {
		vec4 Result = vec4(0, 0, 0, 0);
		for (int i = 0; i < maxiter; i++) {

//vec4 intensity = vec4(texture(colorTexture, point).w);
vec4 intensity = vec4(calculateVoxel(point, vec3(128.0)));
			
			if (intensity.x > Result.x) {
				Result = vec4(intensity.x);
				
				if (transfer == 1) {
					vec4 transferred = texture(transferTexture, (Result.x));
					Result.xyz = transferred.xyz;
				}
			
				if (light == 1 || adaptive == 1) {
					N = calculateGradients(point, dz);
					V = vec3(exit_point - entry_point);
					Result.xyz = lightening(normalize(N), V, Result.xyz);
				}
			}
			if(adaptive == 1){
				point += (1.0/length(N)/80.0) * V;
			}else{			
				point += diff;
			}
			if (Result.w >= 1.0) {
				break;
			}
				
		}
		return Result;
	}
}

vec4 averageOrCompositing(float dz, vec4 entry_point, vec4 exit_point, int maxiter, vec3 diff)
{
	vec3 point = entry_point.xyz;
	vec3 N;
	vec3 V;
	
	float dist = distance(entry_point.xyz, exit_point.xyz);
	
	if (transfer == 1) {
		if (entry_point.w == 0.0) {
			discard; //alpha = 0 man sieht nichts!
		} else {
			vec4 Result = vec4(0, 0, 0, 0);
			for (int i = 0; i < maxiter; i++) {

//vec4 intensity = vec4(texture(colorTexture, point).w);
vec4 intensity = vec4(calculateVoxel(point, vec3(256.0)));
				vec4 transferred = texture(transferTexture, (intensity.x));
				
				if (light == 1 || adaptive == 1) {
					N = calculateGradients(point, dz);
					V = vec3(exit_point - entry_point);
					transferred.xyz = lightening(normalize(N), V, transferred.xyz);
				}
				
				if (average == 1) {
					Result.xyz += transferred.w * transferred.xyz;
					Result.w += transferred.w;
				} else {
					//compositing
					Result.xyz += ((1.0 - Result.w) * transferred.w) * transferred.xyz;
					Result.w += (1.0 - Result.w) * transferred.w;
					if (Result.w >= 1.0) {
						break;
					}
				}
				if(adaptive == 1){
					point += (1.0/length(N)/80.0) * V;
				}else{
					point += diff;
				}
			}
			if (average == 1) {
				Result /= float(maxiter);
				Result.w = 1.0;
			}
			return Result;
		}
	} else { //keine transfer funktion!
		if (entry_point.w == 0.0) {
			discard; //alpha = 0 man sieht nichts!
		} else {
			vec4 Result = vec4(0, 0, 0, 0);
			for (int i = 0; i < maxiter; i++) {

//vec4 intensity = vec4(texture(colorTexture, point).w);
vec4 intensity = vec4(calculateVoxel(point, vec3(256.0)));
				
				if (light == 1 || adaptive == 1) {
					N = calculateGradients(point, dz);
					V = vec3(exit_point - entry_point);
					intensity.xyz = lightening(normalize(N), V, intensity.xyz);
				}
				
				if (average == 1) {
					Result.xyz += intensity.w * intensity.xyz;
					Result.w += intensity.w;
				} else {
					//compositing
					Result.xyz += ((1.0 - Result.w) * intensity.w) * intensity.xyz;
					Result.w += (1.0 - Result.w) * intensity.w;
				}
				if(adaptive == 1){
					if (length(N) != 0)
					{
						point += (1.0/length(N)/80.0) * V;
					} else
					{
						point += (1.0/0.01/80.0) * V;
					}
				}else{
					point += diff;
				}
			}
			if (average == 1) {
				Result /= float(maxiter);
				Result.w = 1.0;
			}
			return Result;
		}
	}
}


void main()
{
	float dz = stepSize; //Schrittgre = 1/200
	vec4 entry_point = texture(entryTexture, fragTexCoordVolume);
	vec4 exit_point = texture(exitTexture, fragTexCoordVolume);
	float iterations = distance(entry_point, exit_point)/dz;
	vec3 diff = (exit_point.xyz - entry_point.xyz)/iterations;
	int maxiter = int(floor(iterations));

	if (mip == 1) 
	{
		out_color = maximumIntensity(dz, entry_point, exit_point, maxiter, diff);
	} 
	else 
	{
		out_color = averageOrCompositing(dz, entry_point, exit_point, maxiter, diff);
	}
}