#version 330 core


layout (points) in;
layout (triangle_strip, max_vertices = 18) out;
    
uniform mat4 perspective;
uniform mat4 view;
uniform mat4 model;
uniform sampler1D transfertex;
uniform vec3 volumeSize;

in VertexData {
    vec4 ex_color;
} VertexIn[1];


out VertexData {
    vec4 ex_color;
} VertexOut;

vec4 check(vec4 tmp)
{
	return min(vec4(1.0), max(vec4(0.0), tmp));
}

void main()                                                   
{
	vec3 grid = volumeSize;
	vec4 x = vec4(16.0/grid.x, 0.0, 0.0,0.0);
	vec4 y = vec4(0.0, 16.0/grid.y, 0.0,0.0);
	vec4 z = vec4(0.0, 0.0, 16.0/grid.z,0.0);
	
	vec4 pos = gl_in[0].gl_Position;
	mat4 transformation = perspective * view * model;	
                                                             
	// output Box
	//Back

	float minx = texture(transfertex, VertexIn[0].ex_color.y).w;
	float maxx = texture(transfertex, VertexIn[0].ex_color.z).w;

	if((minx + maxx) > 0.01 || (maxx-minx)>0.01)
	{
		vec4 p = check(pos+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+x+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+y+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
    		EmitVertex();

		p = check(pos+x+y+z);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos+y);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos+x+y);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos+x);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos+z);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		p = check(pos+x+z);
		VertexOut.ex_color = p;
		gl_Position =  transformation * p;
		EmitVertex();

		EndPrimitive();

		//LEFT SIDE
		p = check(pos);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+y);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
    		EmitVertex();

		p = check(pos+y+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
    		EmitVertex();

		EndPrimitive();

		//RIGHT SIDE
		p = check(pos+x+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+x);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
		EmitVertex();

		p = check(pos+x+y+z);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
    		EmitVertex();

		p = check(pos+x+y);
		VertexOut.ex_color = p;
		gl_Position = transformation * p;
    		EmitVertex();

		EndPrimitive();
		
	}	
}  